
import tensorflow as tf

from model import GAN
from utils import image_utils

FLAGS = tf.flags.FLAGS

# FLAGS Used to export Graph
tf.flags.DEFINE_string('checkpoint_dir', './checkpoints/20181219-0035', 'checkpoints directory path')
tf.flags.DEFINE_string('model', 'model.pb', 'model name')




# Utilized data
tf.flags.DEFINE_string('pretrained_folder', '', 'path of the folder for pretraied encoder models')
tf.flags.DEFINE_string('training_file', '','tfrecords file for training')
tf.flags.DEFINE_string('validation_file', '','tfrecords file for validation')
tf.flags.DEFINE_string('unsupervised_file', '', 'tfrecords file for unsupervised learning')

# Data format
tf.flags.DEFINE_integer('batch_size', 1, 'batch size')
tf.flags.DEFINE_integer('validation_batch_size', 1, ' validation batch size default: 1449')
tf.flags.DEFINE_integer('image_height', 375, 'image size, default: 256')
tf.flags.DEFINE_integer('image_width', 750, 'image size, default: 256')
tf.flags.DEFINE_integer('num_classes', 16, 'number of output classes')

# Training parameters
tf.flags.DEFINE_string('train_type', 'semi-supervised',
                       '[supervised, adversarial, semi-supervised, domain-adaptation] type of training to use')
tf.flags.DEFINE_integer('training_steps', 40000, 'number of training steps')
tf.flags.DEFINE_integer('logging_steps_interval', 1, 'interval of steps between each log')
tf.flags.DEFINE_integer('validation_steps_interval', 1, 'interval of steps between each validation')
tf.flags.DEFINE_integer('save_steps_interval', 10000, 'interval of steps between each log')
tf.flags.DEFINE_integer('semi_start', 0, 'number of steps before start computing unsupervised loss (L-semi)')
tf.flags.DEFINE_boolean('separate_D_training', False,
                        'specifies if the discriminator must be trained in a separate time for real and fake data')
tf.flags.DEFINE_boolean('save_summary_images', False,
                        'specifies if the training images must be saved during the training (NB: higher memory usage)')

# Checkpoint folder
tf.flags.DEFINE_string('load_model',
                       None,
                       'folder of saved model that you wish to continue training (checkpoint/{})')

# Network optimization parameters
tf.flags.DEFINE_float('lambda_adv_sup', 0.01, 'weight for discriminator in supervised mode')
tf.flags.DEFINE_float('lambda_adv_unsup', 0.001, 'weight for discriminator in unsupervised mode')
tf.flags.DEFINE_float('lambda_unsup', 0.1, 'weight for unsupervised loss')
tf.flags.DEFINE_float('threshold_unsup', 0.2, 'threshold applied to confidence map produced by discriminator')

# Generator optimization parameter (generator uses SGD optimizer)
tf.flags.DEFINE_float('generator_learning_rate', 2.5e-4, 'initial learning rate for generator')
tf.flags.DEFINE_float('generator_end_learning_rate', 0.0, 'final decay rate for generator optimizer')
tf.flags.DEFINE_integer('generator_start_decay_step', 0, 'number of steps after which start decay for generator')
tf.flags.DEFINE_integer('generator_decay_steps', 40000, 'number of steps of decay learning rate for generator')
tf.flags.DEFINE_float('generator_decay_power', 0.9, 'power of weight decay for generator optimizer')
tf.flags.DEFINE_float('generator_momentum', 0.9, 'momentum for generator optimizer')
tf.flags.DEFINE_float('generator_weight_decay', 1e-4, 'weight decay for momentum of generator optimizer')

# Discriminator optimization parameter (discriminator uses Adam optimizer)
tf.flags.DEFINE_float('discriminator_learning_rate', 1e-4, 'initial learning rate for discriminator')
tf.flags.DEFINE_float('discriminator_end_learning_rate', 0.0, 'final decay rate for discriminator optimizer')
tf.flags.DEFINE_integer('discriminator_start_decay_step', 0,
                        'number of steps after which start decay for discriminator')
tf.flags.DEFINE_integer('discriminator_decay_steps', 40000, 'number of steps of decay learning rate for discriminator')
tf.flags.DEFINE_float('discriminator_decay_power', 0.9, 'power of weight decay for discriminator optimizer')
tf.flags.DEFINE_float('discriminator_momentum', 0.9, 'momentum for generator optimizer (utilized as beta1 parameter)')
tf.flags.DEFINE_float('discriminator_weight_decay', 1e-4, 'weight decay for discriminator optimizer')


class_weights_GTA_Train = [0.637960276015760, 0.906876401511850, 0.808888109204147, 0.979417971165853,
                                   0.992849926507391, 0.988065518323295, 0.998489616496333, 0.999064071693913,
                                   0.914391166506145, 0.975514686906170, 0.848631300794846, 0.995948128310475,
                                   0.999638046230835, 0.971915722873151, 0.987760747283702, 0.995798327177682,
                                   0.999222562735153, 0.999635347188153, 0.999932073075146]

def export_graph(model_name):
    graph = tf.Graph()

    with graph.as_default():
        tf.set_random_seed(12345)
        gan = GAN(
            train_file=FLAGS.training_file,
            validation_file=FLAGS.validation_file,
            unsupervised_file=FLAGS.unsupervised_file,
            batch_size=FLAGS.batch_size,
            image_height=FLAGS.image_height,
            image_width=FLAGS.image_width,
            num_classes=FLAGS.num_classes,
            pretrained_folder=FLAGS.pretrained_folder,
            lambda_adv_sup=FLAGS.lambda_adv_sup,
            lambda_adv_unsup=FLAGS.lambda_adv_unsup,
            lambda_adv_unsup_d2=FLAGS.lambda_adv_unsup_d2,
            lambda_unsup=FLAGS.lambda_unsup,
            threshold_unsup=FLAGS.threshold_unsup,
            percentili=FLAGS.percentili,
            generator_learning_rate=FLAGS.generator_learning_rate,
            generator_momentum=FLAGS.generator_momentum,
            generator_weight_decay=FLAGS.generator_weight_decay,
            generator_decay_power=FLAGS.generator_decay_power,
            generator_decay_steps=FLAGS.generator_decay_steps,
            generator_start_decay_step=FLAGS.generator_start_decay_step,
            generator_end_learning_rate=FLAGS.generator_end_learning_rate,
            discriminator_learning_rate=FLAGS.discriminator_learning_rate,
            discriminator_momentum=FLAGS.discriminator_momentum,
            discriminator_weight_decay=FLAGS.discriminator_weight_decay,
            discriminator_decay_power=FLAGS.discriminator_decay_power,
            discriminator_decay_steps=FLAGS.discriminator_decay_steps,
            discriminator_start_decay_step=FLAGS.discriminator_start_decay_step,
            discriminator_end_learning_rate=FLAGS.discriminator_end_learning_rate,
            discriminator_learning_rate_2=FLAGS.discriminator_learning_rate_2,
            discriminator_end_learning_rate_2=FLAGS.discriminator_end_learning_rate_2,
            train_type=FLAGS.train_type,
            use_lsgan=FLAGS.use_lsgan,
            save_summary_images=FLAGS.save_summary_images,
            save_histogram_images=FLAGS.save_histogram_images,
            class_weights=class_weights_GTA_Train)

        image_data = tf.placeholder(tf.string, name='input_image')
        input_image = tf.image.decode_jpeg(image_data, channels=3)
        input_image = tf.image.resize_images(input_image, size=(FLAGS.image_size, FLAGS.image_size))
        input_image = image_utils.convert2float(input_image)
        input_image.set_shape([FLAGS.image_size, FLAGS.image_size, 3])
        output_image = gan.G.sample(tf.expand_dims(input_image, 0))

        output_image = tf.identity(output_image, name='output_image')
        restore_saver = tf.train.Saver()

    with tf.Session(graph=graph) as sess:
        sess.run(tf.global_variables_initializer())
        latest_ckpt = tf.train.latest_checkpoint(FLAGS.checkpoint_dir)
        restore_saver.restore(sess, latest_ckpt)
        output_graph_def = tf.graph_util.convert_variables_to_constants(
            sess, graph.as_graph_def(), [output_image.op.name])

        tf.train.write_graph(output_graph_def, 'output_model', model_name, as_text=False)

def main(unused_argv):
    print('Export model...')
    export_graph(FLAGS.model)


if __name__ == '__main__':
    tf.app.run()
