import logging

import numpy as np
import tensorflow as tf

from model import GAN
from utils import tensor_utils

FLAGS = tf.flags.FLAGS
# Checkpoint folder
tf.flags.DEFINE_string('load_model', '', 'folder of saved model that you wish to continue training (checkpoint/{})')

# Utilized data
tf.flags.DEFINE_string('pretrained_folder', 'data/pretrained', 'path of the folder for pretraied encoder models')
tf.flags.DEFINE_string('unsupervised_file', 'data/tfrecords/GTA/Cityscapes_mapped.tfrecords', 'tfrecords file for unsupervised learning')
tf.flags.DEFINE_string('training_file', 'data/tfrecords/GTA/train_div_2.tfrecords','tfrecords file for training')
tf.flags.DEFINE_string('validation_file', 'data/tfrecords/GTA/Cityscapes_mapped_val.tfrecords','tfrecords file for validation')

# Unused parameters for test
# Data format
tf.flags.DEFINE_integer('batch_size', 1, 'batch size')
tf.flags.DEFINE_integer('validation_batch_size', 500, ' validation batch size default: 1449')
tf.flags.DEFINE_integer('image_height', 375, 'image size, default: 256')
tf.flags.DEFINE_integer('image_width', 750, 'image size, default: 256')
tf.flags.DEFINE_integer('num_classes', 19, 'number of output classes')

# Training parameters
tf.flags.DEFINE_string('train_type', 'semi-supervised',
                       '[supervised, adversarial, semi-supervised, after-supervised] type of training to use')
tf.flags.DEFINE_integer('training_steps', 20000, 'number of training steps')
tf.flags.DEFINE_integer('logging_steps_interval', 100, 'interval of steps between each log')
tf.flags.DEFINE_integer('validation_steps_interval', 1000, 'interval of steps between each validation')
tf.flags.DEFINE_integer('save_steps_interval', 10000, 'interval of steps between each log')
tf.flags.DEFINE_integer('semi_start', 5000, 'number of steps before start computing unsupervised loss (L-semi)')
tf.flags.DEFINE_boolean('use_lsgan', False,
                        'define the type of loss of the discriminator (least square or cross_entropy)')

# Network optimization parameters
tf.flags.DEFINE_float('lambda_adv_sup', 0.01, 'weight for discriminator in supervised mode')
tf.flags.DEFINE_float('lambda_adv_unsup', 0.001, 'weight for discriminator in unsupervised mode')
tf.flags.DEFINE_float('lambda_unsup', 0.1, 'weight for discriminator')
tf.flags.DEFINE_float('threshold_unsup', 0.2, 'threshold applied to confidence map produced by discriminator')

# Generator optimization parameter (generator uses SGD optimizer)
tf.flags.DEFINE_float('generator_learning_rate', 1e-4, 'initial learning rate for generator')
tf.flags.DEFINE_float('generator_end_learning_rate', 1e-6, 'final decay rate for generator optimizer')
tf.flags.DEFINE_integer('generator_start_decay_step', 0, 'number of steps after which start decay for generator')
tf.flags.DEFINE_integer('generator_decay_steps', 20000, 'number of steps of decay learning rate for generator')
tf.flags.DEFINE_float('generator_decay_power', 0.9, 'power of weight decay for generator optimizer')
tf.flags.DEFINE_float('generator_momentum', 0.9, 'momentum for generator optimizer')
tf.flags.DEFINE_float('generator_weight_decay', 1e-4, 'weight decay for momentum of generator optimizer')

# Discriminator optimization parameter (discriminator uses Adam optimizer)
tf.flags.DEFINE_float('discriminator_learning_rate', 1e-4, 'initial learning rate for discriminator')
tf.flags.DEFINE_float('discriminator_end_learning_rate', 1e-6, 'final decay rate for discriminator optimizer')
tf.flags.DEFINE_integer('discriminator_start_decay_step', 0,
                        'number of steps after which start decay for discriminator')
tf.flags.DEFINE_integer('discriminator_decay_steps', 20000, 'number of steps of decay learning rate for discriminator')
tf.flags.DEFINE_float('discriminator_decay_power', 0.9, 'power of weight decay for discriminator optimizer')
tf.flags.DEFINE_float('discriminator_momentum', 0.9, 'momentum for generator optimizer (utilized as beta1 parameter)')
tf.flags.DEFINE_float('discriminator_weight_decay', 1e-4, 'weight decay for discriminator optimizer')

#Discriminator optimization paramter
tf.flags.DEFINE_float('discriminator_learning_rate_2',1e-4,'initial lr disc 2')
tf.flags.DEFINE_float('discriminator_end_learning_rate_2',1e-5,'final lr disc 2')



def test():
    # Class weights computed on the SYNTHIA dataset

    #gta
    class_weights_GTA_Train = [0.637960276015760, 0.906876401511850, 0.808888109204147, 0.979417971165853,
                         0.992849926507391, 0.988065518323295, 0.998489616496333, 0.999064071693913,
                         0.914391166506145, 0.975514686906170, 0.848631300794846, 0.995948128310475,
                         0.999638046230835, 0.971915722873151, 0.987760747283702, 0.995798327177682,
                         0.999222562735153, 0.999635347188153, 0.999932073075146]

    # Prevent error
    assert FLAGS.train_type == 'supervised' or FLAGS.train_type == 'adversarial' or FLAGS.train_type == 'semi-supervised' or FLAGS.train_type == 'domain-adaptation', 'value not recognised! Allowed value are: [supervised, adversarial, semi-supervised]'
    assert FLAGS.load_model is not None, 'Needed pre-trained model for test!'

    #checkpoints_dir = "checkpoints/" + FLAGS.load_model.lstrip("checkpoints/")
    checkpoints_dir = FLAGS.load_model
    graph = tf.Graph()
    with graph.as_default():
        tf.set_random_seed(12345)
        gan = GAN(
            train_file=FLAGS.training_file,
            validation_file=FLAGS.validation_file,
            unsupervised_file=FLAGS.unsupervised_file,
            batch_size=FLAGS.batch_size,
            image_height=FLAGS.image_height,
            image_width=FLAGS.image_width,
            num_classes=FLAGS.num_classes,
            pretrained_folder=FLAGS.pretrained_folder,
            lambda_adv_sup=FLAGS.lambda_adv_sup,
            lambda_adv_unsup=FLAGS.lambda_adv_unsup,
            lambda_adv_unsup_d2=FLAGS.lambda_adv_unsup_d2,
            lambda_unsup=FLAGS.lambda_unsup,
            threshold_unsup=FLAGS.threshold_unsup,
            percentili=FLAGS.percentili,
            generator_learning_rate=FLAGS.generator_learning_rate,
            generator_momentum=FLAGS.generator_momentum,
            generator_weight_decay=FLAGS.generator_weight_decay,
            generator_decay_power=FLAGS.generator_decay_power,
            generator_decay_steps=FLAGS.generator_decay_steps,
            generator_start_decay_step=FLAGS.generator_start_decay_step,
            generator_end_learning_rate=FLAGS.generator_end_learning_rate,
            discriminator_learning_rate=FLAGS.discriminator_learning_rate,
            discriminator_momentum=FLAGS.discriminator_momentum,
            discriminator_weight_decay=FLAGS.discriminator_weight_decay,
            discriminator_decay_power=FLAGS.discriminator_decay_power,
            discriminator_decay_steps=FLAGS.discriminator_decay_steps,
            discriminator_start_decay_step=FLAGS.discriminator_start_decay_step,
            discriminator_end_learning_rate=FLAGS.discriminator_end_learning_rate,
            discriminator_learning_rate_2=FLAGS.discriminator_learning_rate_2,
            discriminator_end_learning_rate_2=FLAGS.discriminator_end_learning_rate_2,
            train_type=FLAGS.train_type,
            use_lsgan=FLAGS.use_lsgan,
            save_summary_images=FLAGS.save_summary_images,
            save_histogram_images=FLAGS.save_histogram_images,
            class_weights=class_weights_GTA_Train)

        #G_sup_loss, G_composed_loss_adv, G_composed_loss_adv_semi, D_loss, G_optimizer_sup, G_optimizer_unsup_adv_semi, D_optimizer, training_summary_sup, training_summary_unsup = gan.training_model()
        #G_sup_loss, G_composed_loss_adv, G_composed_loss_adv_semi, D_loss, G_optimizer_unsup_adv_semi, training_summary_sup, training_summary_unsup = gan.training_model()
        iterator, mean_validation_loss, mean_validation_update_op, confusion_matrix, accuracy, accuracy_update_op, mean_IoU, mean_IoU_update_op, validation_summary = gan.validation_model()

        train_writer = tf.summary.FileWriter(checkpoints_dir, graph)
        saver = tf.train.Saver()
    with tf.Session(graph=graph, config=None) as sess:
        tf.set_random_seed(12345)

        # Load model
        sess.run(tf.local_variables_initializer())
        sess.run(tf.global_variables_initializer())
        checkpoint = tf.train.get_checkpoint_state(checkpoints_dir)
        meta_graph_path = checkpoint.model_checkpoint_path + ".meta"
        if FLAGS.train_type == 'domain-adaptation':  # load only the needed components of the networks
        #if FLAGS.train_type == 'supervised':  # load only the needed components of the networks
            restore_variables = gan.G.all_var + gan.D.variables
            loader = tf.train.Saver(var_list=restore_variables, reshape=True)
            loader.restore(sess, tf.train.latest_checkpoint(checkpoints_dir))
        else:
            loader = tf.train.import_meta_graph(meta_graph_path)
            loader.restore(sess, tf.train.latest_checkpoint(checkpoints_dir))
        start_step = int(meta_graph_path.split("-")[-1].split(".")[0])
        step = start_step

        coord = tf.train.Coordinator()
        threads = tf.train.start_queue_runners(sess=sess, coord=coord)
        try:
            # Compute validation metrics
            sess.run(tf.local_variables_initializer())
            sess.run([iterator.initializer])
            n = 0
            local_confusion_matrix = np.zeros((FLAGS.num_classes, FLAGS.num_classes), dtype=np.int)
            for i in range(FLAGS.validation_batch_size):
                try:
                    cm, _, _, _ = sess.run(
                        [confusion_matrix, mean_validation_update_op, mean_IoU_update_op, accuracy_update_op])
                    local_confusion_matrix += cm
                    n += 1
                except tf.errors.OutOfRangeError:  # no more items
                    logging.info(
                        ' The chosen value ({}) exceed number of validation images! Used {} instead.'.format(
                            FLAGS.validation_batch_size, n))
                    break
                print(str(i + 1) + '/' + str(FLAGS.validation_batch_size))
            mean_validation_loss_val, mean_IoU_val, accuracy_val, summary = sess.run(
                [mean_validation_loss, mean_IoU, accuracy, validation_summary])
            train_writer.add_summary(summary, step)
            train_writer.flush()
            logging.info(' Validation_loss : {}'.format(mean_validation_loss_val))
            logging.info(' accuracy : {:0.2f}%'.format(accuracy_val * 100))
            #tensor_utils.compute_and_print_IoU_per_class(local_confusion_matrix, FLAGS.num_classes)
            tensor_utils.compute_and_print_IoU_per_class_and_acc_perclass(local_confusion_matrix, FLAGS.num_classes)
        except KeyboardInterrupt:
            logging.info('Interrupted')
            coord.request_stop()
        except Exception as e:
            coord.request_stop(e)
        finally:
            coord.request_stop()
            coord.join(threads)

def main(unused_argv):
    test()


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    tf.app.run()
