import logging
import os
from datetime import datetime

import numpy as np
import tensorflow as tf

from model import GAN
from utils import tensor_utils

FLAGS = tf.flags.FLAGS

# Utilized data
tf.flags.DEFINE_string('pretrained_folder', '../../../Pretrained/deeplabv2', 'path of the folder for pretraied encoder models')
tf.flags.DEFINE_string('training_file', '../../../Datasets/GTA/tfrecords/train_div_2.tfrecords','tfrecords file for training')
tf.flags.DEFINE_string('validation_file', '../../../Datasets/Cityscapes/tfrecords/Cityscapes_mapped_val.tfrecords','tfrecords file for validation')
tf.flags.DEFINE_string('unsupervised_file', '../../../Datasets/Cityscapes/tfrecords/Cityscapes_mapped.tfrecords', 'tfrecords file for unsupervised learning')


# Data format
tf.flags.DEFINE_integer('batch_size', 1, 'batch size')
tf.flags.DEFINE_integer('validation_batch_size', 500, ' validation batch size default: 500')
tf.flags.DEFINE_integer('image_height', 375, 'image size, default: 375')
tf.flags.DEFINE_integer('image_width', 750, 'image size, default: 750')
tf.flags.DEFINE_integer('num_classes', 19, 'number of output classes')

# Training parameters
tf.flags.DEFINE_string('train_type', 'semi-supervised',
                       '[supervised, adversarial, semi-supervised, domain-adaptation] type of training to use')
tf.flags.DEFINE_integer('training_steps', 20000, 'number of training steps')
tf.flags.DEFINE_integer('logging_steps_interval', 2000, 'interval of steps between each log')
tf.flags.DEFINE_integer('validation_steps_interval', 4000, 'interval of steps between each validation')
tf.flags.DEFINE_integer('save_steps_interval', 20000, 'interval of steps between each save')
tf.flags.DEFINE_integer('semi_start', 5000, 'number of steps before start computing unsupervised loss (L-semi)')
tf.flags.DEFINE_boolean('use_lsgan', False,
                        'define the type of loss of the discriminator (least square or cross_entropy)')
tf.flags.DEFINE_boolean('save_summary_images', False,
                        'specifies if the training images must be saved during the training (NB: higher memory usage)')
tf.flags.DEFINE_boolean('save_histogram_images', False,
                        'specifies if the histogram must be saved during the training (debug)')

# Checkpoint folder
tf.flags.DEFINE_string('load_model',
                       None,
                       'folder of saved model that you wish to continue training (checkpoint/{})')

# Network optimization parameters
tf.flags.DEFINE_float('lambda_adv_sup', 0.01, 'weight for discriminator in supervised mode')
tf.flags.DEFINE_float('lambda_adv_unsup', 0.001, 'weight for discriminator in unsupervised mode')
tf.flags.DEFINE_float('lambda_adv_unsup_d2', 0.001, 'weight for discriminator in unsupervised mode')
tf.flags.DEFINE_float('lambda_unsup', 0.1, 'weight for unsupervised loss')
tf.flags.DEFINE_float('threshold_unsup', 0.2, 'threshold applied to confidence map produced by discriminator')
tf.flags.DEFINE_float('percentili', 0.0, 'percentile per threshold')

# Generator optimization parameter (generator uses Adam optimizer)
tf.flags.DEFINE_float('generator_learning_rate', 1e-4, 'initial learning rate for generator')
tf.flags.DEFINE_float('generator_end_learning_rate', 1e-6, 'final decay rate for generator optimizer')
tf.flags.DEFINE_integer('generator_start_decay_step', 0, 'number of steps after which start decay for generator')
tf.flags.DEFINE_integer('generator_decay_steps', 20000, 'number of steps of decay learning rate for generator')
tf.flags.DEFINE_float('generator_decay_power', 0.9, 'power of weight decay for generator optimizer')
tf.flags.DEFINE_float('generator_momentum', 0.9, 'beta  for generator optimizer')
tf.flags.DEFINE_float('generator_weight_decay', 1e-4, 'weight decay for momentum of generator optimizer')

# Discriminator optimization parameter (discriminator uses Adam optimizer)
tf.flags.DEFINE_float('discriminator_learning_rate', 1e-4, 'initial learning rate for discriminator')
tf.flags.DEFINE_float('discriminator_end_learning_rate', 1e-6, 'final decay rate for discriminator optimizer')
tf.flags.DEFINE_integer('discriminator_start_decay_step', 0,
                        'number of steps after which start decay for discriminator')
tf.flags.DEFINE_integer('discriminator_decay_steps', 20000, 'number of steps of decay learning rate for discriminator')
tf.flags.DEFINE_float('discriminator_decay_power', 0.9, 'power of weight decay for discriminator optimizer')
tf.flags.DEFINE_float('discriminator_momentum', 0.9, 'momentum for generator optimizer (utilized as beta1 parameter)')
tf.flags.DEFINE_float('discriminator_weight_decay', 1e-4, 'weight decay for discriminator optimizer')

#Discriminator optimization paramter
tf.flags.DEFINE_float('discriminator_learning_rate_2',1e-4,'initial lr disc 2')
tf.flags.DEFINE_float('discriminator_end_learning_rate_2',1e-4,'final lr disc 2')



def train():
    # Class weights computed on the GTA dataset


    class_weights_GTA_Train = [0.637960276015760, 0.906876401511850, 0.808888109204147, 0.979417971165853,
                                   0.992849926507391, 0.988065518323295, 0.998489616496333, 0.999064071693913,
                                   0.914391166506145, 0.975514686906170, 0.848631300794846, 0.995948128310475,
                                   0.999638046230835, 0.971915722873151, 0.987760747283702, 0.995798327177682,
                                   0.999222562735153, 0.999635347188153, 0.999932073075146]

    # Prevent errors
    assert FLAGS.train_type == 'supervised' or FLAGS.train_type == 'adversarial' or FLAGS.train_type == 'semi-supervised', 'value not recognised! Allowed value are: [supervised, adversarial, semi-supervised]'

    # Define checkpoint folder name
    if FLAGS.load_model is not None:
        checkpoints_dir = "checkpoints/" + FLAGS.load_model.lstrip("checkpoints/")
    else:
        current_time = datetime.now().strftime("%Y%m%d-%H%M") \
                       + '_semi-start' + str(FLAGS.semi_start) \
                       + '_G-lr-' + str(FLAGS.generator_learning_rate) \
                       + '_D-lr-' + str(FLAGS.discriminator_learning_rate) \
                       + '_yas-' + str(FLAGS.lambda_adv_sup) \
                       + '_yau-' + str(FLAGS.lambda_adv_unsup) \
                       + '_yaud2-' + str(FLAGS.lambda_adv_unsup_d2) \
                       + '_yu-' + str(FLAGS.lambda_unsup) \
                       + '_t-' + str(FLAGS.threshold_unsup) \
                       + '_P' + str(FLAGS.percentili)
        checkpoints_dir = "checkpoints/{}".format(current_time)
        try:
            os.makedirs(checkpoints_dir)
        except os.error:
            pass

    # Define the model
    graph = tf.Graph()
    with graph.as_default():
        tf.set_random_seed(12345)
        gan = GAN(
            train_file=FLAGS.training_file,
            validation_file=FLAGS.validation_file,
            unsupervised_file=FLAGS.unsupervised_file,
            batch_size=FLAGS.batch_size,
            image_height=FLAGS.image_height,
            image_width=FLAGS.image_width,
            num_classes=FLAGS.num_classes,
            pretrained_folder=FLAGS.pretrained_folder,
            lambda_adv_sup=FLAGS.lambda_adv_sup,
            lambda_adv_unsup=FLAGS.lambda_adv_unsup,
            lambda_adv_unsup_d2=FLAGS.lambda_adv_unsup_d2,
            lambda_unsup=FLAGS.lambda_unsup,
            threshold_unsup=FLAGS.threshold_unsup,
            percentili=FLAGS.percentili,
            generator_learning_rate=FLAGS.generator_learning_rate,
            generator_momentum=FLAGS.generator_momentum,
            generator_weight_decay=FLAGS.generator_weight_decay,
            generator_decay_power=FLAGS.generator_decay_power,
            generator_decay_steps=FLAGS.generator_decay_steps,
            generator_start_decay_step=FLAGS.generator_start_decay_step,
            generator_end_learning_rate=FLAGS.generator_end_learning_rate,
            discriminator_learning_rate=FLAGS.discriminator_learning_rate,
            discriminator_momentum=FLAGS.discriminator_momentum,
            discriminator_weight_decay=FLAGS.discriminator_weight_decay,
            discriminator_decay_power=FLAGS.discriminator_decay_power,
            discriminator_decay_steps=FLAGS.discriminator_decay_steps,
            discriminator_start_decay_step=FLAGS.discriminator_start_decay_step,
            discriminator_end_learning_rate=FLAGS.discriminator_end_learning_rate,
            discriminator_learning_rate_2=FLAGS.discriminator_learning_rate_2,
            discriminator_end_learning_rate_2=FLAGS.discriminator_end_learning_rate_2,
            train_type=FLAGS.train_type,
            use_lsgan=FLAGS.use_lsgan,
            save_summary_images=FLAGS.save_summary_images,
            save_histogram_images=FLAGS.save_histogram_images,
            class_weights=class_weights_GTA_Train)

        G_sup_loss, G_composed_loss_adv, G_composed_loss_adv_semi, D_loss, D2_loss, G_optimizer_sup, G_optimizer_unsup_adv_semi, D_optimizer, D2_optimizer, training_summary_sup, training_summary_unsup,debug_threshold = gan.training_model()
        iterator, mean_validation_loss, mean_validation_update_op, confusion_matrix, accuracy, accuracy_update_op, mean_IoU, mean_IoU_update_op, validation_summary = gan.validation_model()
        train_writer = tf.summary.FileWriter(checkpoints_dir, graph)
        saver = tf.train.Saver(max_to_keep=None)
    summary = None
    G_loss_val = 0
    D_1_loss_val = 0
    D_2_loss_val = 0
    step = 0
    media_th = 0
    count_class = np.zeros(FLAGS.num_classes)
    file_th = open(checkpoints_dir+ ".txt", "w+")


    logging.info(' training type: {}'.format(FLAGS.train_type))
    logging.info(' least square: {}'.format(FLAGS.use_lsgan))
    logging.info(' discriminator learning rate: {}'.format(FLAGS.discriminator_learning_rate))
    logging.info(' lambda adv sup: {}'.format(FLAGS.lambda_adv_sup))
    logging.info(' lambda adv unsup: {}'.format(FLAGS.lambda_adv_unsup))
    logging.info(' lambda adv unsup su d2: {}'.format(FLAGS.lambda_adv_unsup_d2))
    logging.info(' lambda unsup: {}'.format(FLAGS.lambda_unsup))
    logging.info(' percentile: {}'.format(FLAGS.percentili))


    # Load checkpoint if provided
    with tf.Session(graph=graph, config=None) as sess:
        tf.set_random_seed(12345)
        if FLAGS.load_model is not None:  # Restore pre trained checkpoint
            sess.run(tf.local_variables_initializer())
            sess.run(tf.global_variables_initializer())
            checkpoint = tf.train.get_checkpoint_state(checkpoints_dir)
            meta_graph_path = checkpoint.model_checkpoint_path + ".meta"
            loader = tf.train.import_meta_graph(meta_graph_path)
            loader.restore(sess, tf.train.latest_checkpoint(checkpoints_dir))
            start_step = int(meta_graph_path.split("-")[-1].split(".")[0])
            step = start_step
        else:
            sess.run(tf.global_variables_initializer())
            sess.run(tf.local_variables_initializer())
            loader = tf.train.Saver(var_list=gan.G.encoder_var, reshape=True)
            loader.restore(sess, FLAGS.pretrained_folder + '/deeplab_resnet_init.ckpt')

        coord = tf.train.Coordinator()
        threads = tf.train.start_queue_runners(sess=sess, coord=coord)

        # Train
        try:
            while not coord.should_stop() and step < FLAGS.training_steps:
                # Separate different types of training
                if FLAGS.train_type == 'supervised':
                    _, G_loss_val, summary = (
                        sess.run(
                            [G_optimizer_sup, G_sup_loss, training_summary_sup]
                        )
                    )
                elif FLAGS.train_type == 'adversarial':
                    # Train generator
                    _, G_loss_val, summary = (
                        sess.run(
                            [G_optimizer_sup, G_sup_loss, training_summary_sup]
                        )
                    )
                    train_writer.add_summary(summary, step)
                    train_writer.flush()
                    # Train discriminator
                    _, D_1_loss_val, summary = (
                        sess.run(
                            [D_optimizer, D_loss, training_summary_sup]
                        )
                    )
                elif FLAGS.train_type == 'semi-supervised':
                    # Do semi first (As in the paper)
                    if step < FLAGS.semi_start:
                        # Train only with unsupervised adversarial loss
                        _, G_loss_val, summary_unsup,threshold = (
                            sess.run(
                                [G_optimizer_unsup_adv_semi, G_composed_loss_adv_semi,
                                 training_summary_unsup,debug_threshold], feed_dict={gan.lambda_semi: 0.0}
                            )
                        )
                    else:
                        # Train with unsupervised adversarial loss and unsupervised cross-entropy (L-semi)
                        _, G_loss_val, summary_unsup,threshold = (
                            sess.run(
                                [G_optimizer_unsup_adv_semi, G_composed_loss_adv_semi,
                                 training_summary_unsup,debug_threshold],
                                feed_dict={gan.lambda_semi: FLAGS.lambda_unsup}
                            )
                        )
                    train_writer.add_summary(summary_unsup, step)
                    train_writer.flush()
                    _, _, D_1_loss_val, D_2_loss_val, summary = (
                        sess.run(
                            [D_optimizer, D2_optimizer, D_loss, D2_loss, training_summary_sup]
                        )
                    )

                train_writer.add_summary(summary, step)
                train_writer.flush()


                # Output computed metrics
                if step % FLAGS.logging_steps_interval == 0:
                    logging.info('-----------Step %d:-------------' % step)
                    if FLAGS.train_type == 'supervised':
                        logging.info(' G_loss : {}'.format(G_loss_val))

                    elif FLAGS.train_type == 'adversarial' or FLAGS.train_type == 'semi-supervised' or FLAGS.train_type == 'domain-adaptation':
                        logging.info(' G_loss : {}'.format(G_loss_val))
                        logging.info(' D_loss : {}'.format(D_1_loss_val))
                        logging.info(' D2_loss : {}'.format(D_2_loss_val))
                        logging.info(' Threshold : {}'.format(threshold))

                # Compute validation metrics
                if step % FLAGS.validation_steps_interval == 0: #and not step ==0:
                    sess.run(tf.local_variables_initializer())
                    sess.run([iterator.initializer])
                    n = 0
                    local_confusion_matrix = np.zeros((FLAGS.num_classes, FLAGS.num_classes), dtype=np.int)
                    for i in range(FLAGS.validation_batch_size):
                        try:
                            cm, _, _, _ = sess.run(
                                [confusion_matrix, mean_validation_update_op, mean_IoU_update_op, accuracy_update_op])
                            local_confusion_matrix += cm
                            n += 1
                        except tf.errors.OutOfRangeError:  # no more items
                            logging.info(
                                ' The chosen value ({}) exceed number of validation images! Used {} instead.'.format(
                                    FLAGS.validation_batch_size, n))
                            break
                    mean_validation_loss_val, mean_IoU_val, accuracy_val, summary = sess.run(
                        [mean_validation_loss, mean_IoU, accuracy, validation_summary])
                    train_writer.add_summary(summary, step)
                    train_writer.flush()
                    logging.info(' Validation_loss : {}'.format(mean_validation_loss_val))
                    logging.info(' accuracy : {:0.2f}%'.format(accuracy_val * 100))
                    tensor_utils.compute_and_print_IoU_per_class(local_confusion_matrix, FLAGS.num_classes)
                    # Save the weight of the network
                    if step % FLAGS.save_steps_interval == 0:
                        save_path = saver.save(sess, checkpoints_dir + "/model.ckpt", global_step=step)
                        logging.info("Model saved in file: %s" % save_path)
                        max_mIoU = mean_IoU_val

                if FLAGS.train_type == 'semi-supervised':
                    for i in range(0,FLAGS.num_classes):
                        if threshold[i] != 0:
                            count_class[i] = count_class[i]+1

                    # save threshold to a file
                    th = threshold.reshape(1, threshold.shape[0])
                    np.savetxt(file_th, th, delimiter='   ')
                    file_th.write("\n")


                    media_th += threshold

                step += 1

        except KeyboardInterrupt:
            logging.info('Interrupted')
            coord.request_stop()
        except Exception as e:
            coord.request_stop(e)
        finally:
            # Output computed metrics
            logging.info('-----------Step %d:-------------' % step)
            if FLAGS.train_type == 'supervised':
                logging.info(' G_loss : {}'.format(G_loss_val))

            elif FLAGS.train_type == 'adversarial' or FLAGS.train_type == 'semi-supervised' or FLAGS.train_type == 'domain-adaptation':
                logging.info(' G_loss : {}'.format(G_loss_val))
                logging.info(' D_loss : {}'.format(D_1_loss_val))
                logging.info(' D2_loss : {}'.format(D_2_loss_val))
                media_th = media_th / count_class
                logging.info(' Media th : {}'.format(media_th))

            # Compute validation metrics
            sess.run(tf.local_variables_initializer())
            sess.run([iterator.initializer])
            n = 0
            local_confusion_matrix = np.zeros((FLAGS.num_classes, FLAGS.num_classes), dtype=np.int)
            for i in range(FLAGS.validation_batch_size):
                try:
                    cm, _, _, _ = sess.run(
                        [confusion_matrix, mean_validation_update_op, mean_IoU_update_op, accuracy_update_op])
                    local_confusion_matrix += cm
                    n += 1
                except tf.errors.OutOfRangeError:  # no more items
                    logging.info(
                        ' The chosen value ({}) exceed number of validation images! Used {} instead.'.format(
                            FLAGS.validation_batch_size, n))
                    break
            mean_validation_loss_val, mean_IoU_val, accuracy_val, summary = sess.run(
                [mean_validation_loss, mean_IoU, accuracy, validation_summary])
            train_writer.add_summary(summary, step)
            train_writer.flush()
            logging.info(' Validation_loss : {}'.format(mean_validation_loss_val))
            logging.info(' accuracy : {:0.2f}%'.format(accuracy_val * 100))
            tensor_utils.compute_and_print_IoU_per_class(local_confusion_matrix, FLAGS.num_classes)
            if FLAGS.train_type != 'domain-adaptation':
                save_path = saver.save(sess, checkpoints_dir + "/model.ckpt", global_step=step)
                logging.info("Model saved in file: %s" % save_path)

            # When done, ask the threads to stop.
            coord.request_stop()
            coord.join(threads)
            file_th.close()


def main(unused_argv):
    train()


if __name__ == '__main__':
    print(os.getcwd())
    os.chdir(os.path.dirname(os.path.realpath(__file__)))
    print(os.getcwd())
    logging.basicConfig(level=logging.INFO)
    tf.app.run()
